use serde::{Serialize, Deserialize};
use wasm_bindgen::prelude::*;
use serde_wasm_bindgen::{to_value, from_value};

pub mod instruction;
use instruction::*;

#[derive(Clone, Debug, PartialEq, Serialize, Deserialize)]
#[wasm_bindgen]
pub struct Node {
    pub x: f32,
    pub y: f32,
}

#[derive(Clone, Debug, Serialize, Deserialize)]
#[wasm_bindgen]
pub struct Program {
    nodes: Vec<Node>,
    instructions: Vec<Instruction>
}

impl Program {
    pub fn new(nodes: Vec<Node>, instructions: Vec<Instruction>) -> Option<Program> {
        if instructions.len() == 0 {
            return None
        }

        Some(Program {
            nodes,
            instructions,
        })
    }

    pub fn nodes(&self) -> &Vec<Node> {
        &self.nodes
    }

    pub fn instructions(&self) -> &Vec<Instruction> {
        &self.instructions
    }
}

#[wasm_bindgen]
extern "C" {
    #[wasm_bindgen(typescript_type = "Node[]")]
    pub type NodeArray;

    #[wasm_bindgen(typescript_type = "Instruction[]")]
    pub type InstructionArray;

    #[wasm_bindgen(typescript_type = "Error")]
    pub type IError;
}

impl IError {
    pub fn from_err<E>(err: E) -> IError where JsValue: From<E> {
        IError::from(JsValue::from(err))
    }
}

#[wasm_bindgen]
impl Program {
    pub fn from(nodes: NodeArray, instructions: InstructionArray) -> Result<Program, IError> {
        let nodes: Vec<Node> = from_value(nodes.into()).map_err(IError::from_err)?;
        let instructions: Vec<Instruction> = from_value(instructions.into()).map_err(IError::from_err)?;

        Program::new(nodes, instructions).ok_or_else(|| {
            unimplemented!();
        })
    }

    pub fn empty() -> Program {
        Self::new(vec![], vec![Instruction::noop()]).unwrap()
    }

    #[wasm_bindgen(getter = nodes)]
    pub fn nodes_js(&self) -> JsValue {
        to_value(&self.nodes).unwrap()
    }

    #[wasm_bindgen(js_name = toJSON)]
    pub fn to_json(&self) -> JsValue {
        to_value(&self).unwrap()
    }
}
